# coding:utf-8

import pymongo
from webmother.db import mongo as mother_db
from tornado.options import options

MongoDB = {
    'active': True,  # 是否启用
    'dev': {
        'host': 'udl.link',
        'port': 27017,
        'db': 'wtdb',
        'user': 'app',
        'pwd': 'Wllt2app'
    },
    'prod': {
        'host': 'udl.link',
        'port': 27017,
        'db': 'wtdb',
        'user': 'app',
        'pwd': 'Wllt2app'
    }
}

# web-mother相关集合
catalog = None

# wallet DB
mongo_client = None
mongo_db = None

wallet = None
contract = None


def init():
    if not MongoDB['active']:
        return

    global catalog

    global mongo_client
    global mongo_db

    global wallet
    global contract

    mongo_cfg = MongoDB[options.env]

    if mongo_client is not None:
        return

    mother_db.init()

    # 引用webmother中的数据集合
    catalog = mother_db.catalog
    # END

    mongo_client = pymongo.MongoClient(host=mongo_cfg['host'],
                                       port=mongo_cfg['port'],
                                       username=mongo_cfg['user'],
                                       password=mongo_cfg['pwd'],
                                       authSource=mongo_cfg['db'])
    mongo_db = mongo_client[mongo_cfg['db']]

    # my collections
    wallet = mongo_db.wallet
    contract = mongo_db.contract
    # END

    # 创建索引
    _wallet_index()
    _contract_index()

    # 初始化系统数据
    _init_data()


def start_session():
    return mongo_client.start_session()


def _wallet_index():
    """
    {
        "_id": ObjectId('5c729df2e155ac16da86a1d0'),
        "uid": ObjectId('5c9c25cae155acde223d5472'),                # 用户ID
        "catalog": ObjectId('5c710622e155ac0c39c8b66d'),            # 所属分类节点ID, 即所属区块链
        "address": '0x58a81beab5f9948114d267ee834d82928accb030',    # 钱包地址
        "contract_id": ObjectId('5c729df2e155ac16da86a1d0'),        # 合约ID，也是币种定义
        "type": "ERC20",     # 币种类型，如：main：主币（本系统将主币也作为一种特殊的合约），ERC20：ERC20币，ERC721：ERC721币
        "alias": '理财钱包',
        "mark": '我不想当韭菜',
        "status": 10,          # -10:已删除，0: 休眠中，10: 活动中
        "created": 1551015331186,
        "updated": 1551015331186
    }
    """
    wallet.create_index('uid')
    wallet.create_index('address')


def _contract_index():
    """
    {
        "_id": ObjectId('5c729df2e155ac16da86a1d0'),
        "catalog": ObjectId('5c710622e155ac0c39c8b66d'), # 所属分类节点ID, 即所属区块链
        "name": 'Simple Token',
        "symbol": 'TST',
        "decimals": 18,       # 有效位长度，即小数位数
        "icon": "https://cdn2.iconfinder.com/data/icons/5/100/cryptocurrency_blockchain_crypto-02-512.png",
        "type": "ERC20",     # 币种类型，如：main：主币（本系统将主币也作为一种特殊的合约），ERC20：ERC20币，ERC721：ERC721币
        "spec": {            # 主币spec为空
            "address": '0x7af963cF6D228E564e2A0aA0DdBF06210B38615D',  # 合约地址
            "data": [...],                                            # 合约相关数据，不同链存在不同的数据
        }
        "status": 10,          # -10:已删除，0: 休眠中，10: 活动中
        "created": 1551015331186,
        "updated": 1551015331186
    }

    本系统将主币也作为一种特殊的合约！！！
    """
    contract.create_index('catalog')
    contract.create_index('spec.address')


def _init_data():
    pass



# coding=utf-8

import webmother
from webmother.service.passport import Passport

from qccwallet.db import mongo
from qccwallet import routes

# 在通证中增加对合约域的权限控制能力
lic_domain = 'contract'
max_lic_text = '{}:111111111;;'.format(lic_domain)
Passport.add_profile(lic_domain, {
    'switch': [
        "create",
        "read",
        "update",
        "remove",
        "submit",
        "audit",
        "reject",
        "activate",
        "deactivate"
    ]
})

Passport.add_display({
    'zh': {
        'contract': '合约管理',
        'contract.switch': '权限开关',
        'contract.switch.create': '创建',
        'contract.switch.read': '读取',
        'contract.switch.remove': '移除',
        'contract.switch.submit': '提交',
        'contract.switch.audit': '审核',
        'contract.switch.reject': '驳回',
        'contract.switch.activate': '上架',
        'contract.switch.deactivate': '下架',
    },
    'en': {
        'contract': 'Contract Manage',
        'contract.switch': 'Switches',
        'contract.switch.create': 'Create',
        'contract.switch.read': 'Read',
        'contract.switch.remove': 'Remove',
        'contract.switch.submit': 'Submit',
        'contract.switch.audit': 'Audit',
        'contract.switch.reject': 'Reject',
        'contract.switch.activate': 'Activate',
        'contract.switch.deactivate': 'Deactivate',
    }
})


def init(app):
    # 初始化webmother
    webmother.init(app)

    # 初始化本系统数据库
    mongo.init()

    # 加载路由模块
    app.load_routes(routes)



from tornado import websocket, gen
import json
from qccwallet.service import ctrl_wallet


class Dispatcher(websocket.WebSocketHandler):
    def data_received(self, chunk):
        print('data_received')

    def check_origin(self, origin):
        print('check_origin: %s' % origin)
        lang = self.request.headers.get("Origin")
        print('Lang: %s' % lang)
        return True

    def open(self):
        print("WebSocket opened")

    @gen.coroutine
    def on_message(self, message):
        print(message)
        self.write_message(message)
        try:
            req = json.loads(message)

            print(req)

            if 'method' in req:
                method = req['method']
                params = req['params']

                if method == 'get_transaction':
                    ret = ctrl_wallet.get_tx_info(*tuple(params))
                else:
                    ret = 'unknown'

                self.write_message(ret)
            else:
                self.write_message(u'Missing method in request')

        except json.JSONDecodeError:
            self.write_message(u'Not valid JSON message')

    def on_close(self):
        print("WebSocket closed")



import config
from tweb.error_exception import ErrException, ERROR
from web3 import Web3
import glob
from eth_account.messages import defunct_hash_message
import logging

provider = Web3.IPCProvider(config.Ethereum['internal']['ipc'])
web3 = Web3(provider)

contracts = dict()


def new_account(pwd):
    return web3.personal.newAccount(pwd).lower()


def is_address(address):
    return web3.isAddress(address)


def get_contract(contract_address, abi):
    try:
        contract_address = to_crc_address(contract_address)
        ret = contracts.get(contract_address)
        if ret is None:
            ret = web3.eth.contract(contract_address, abi=abi)
            contracts[contract_address] = ret
        return ret
    except Exception as e:
        logging.exception(e)
        return None


def to_crc_address(address):
    return web3.toChecksumAddress(address)


def sign_tx(sender_address, tx, password):
    sender_address = web3.toChecksumAddress(sender_address)

    files = glob.glob(r'{}/keystore/*--{}'.format(config.Ethereum['internal']['path'], sender_address.lower()[2:]))
    if len(files) > 0:
        with open(files[0]) as key_file:
            try:
                signed = web3.eth.account.signTransaction(tx, web3.eth.account.decrypt(key_file.read(), password))
            except Exception as e:
                raise ErrException(ERROR.E40000, extra='sign failed, %s' % e)
            return {
                'rawTransaction': signed.rawTransaction.hex(),
                'tx': tx
            }
    else:
        raise ErrException(ERROR.E40000, extra='not found the from address')


def sign_msg(sender_address, msg, password):
    sender_address = web3.toChecksumAddress(sender_address)

    msg_hash = defunct_hash_message(text=msg)

    files = glob.glob(r'{}/keystore/*--{}'.format(config.Ethereum['internal']['path'], sender_address.lower()[2:]))
    if len(files) > 0:
        with open(files[0]) as key_file:
            try:
                signed = web3.eth.account.signHash(msg_hash, web3.eth.account.decrypt(key_file.read(), password))
            except Exception as e:
                raise ErrException(ERROR.E40000, extra='sign failed, addrss and password not match')
            return {
                'rawMessage': signed.messageHash.hex(),
                'message_hash': msg_hash
            }
    else:
        raise ErrException(ERROR.E40000, extra='not found the from address')



# coding=utf-8

from tweb import base_handler, myweb
from tornado import gen
import json
from qccwallet.service.async_wrap import ctrl_wallet


class MyWalletsHandler(base_handler.BaseHandler):

    @myweb.authenticated
    @gen.coroutine
    def get(self):
        user_id = self.request.headers.get('x-user-id')
        access_token = self.request.headers.get('x-access-token')

        ret = yield ctrl_wallet.query_mine(user_id, access_token)
        return self.write({'list': ret})


class WalletHandler(base_handler.BaseHandler):
    """
    钱包账户操作
    """

    @myweb.authenticated
    @gen.coroutine
    def post(self, contract_id, wallet_id):
        user_id = self.request.headers.get('x-user-id')
        access_token = self.request.headers.get('x-access-token')

        data = json.loads(self.request.body.decode('utf-8'))
        ret = yield ctrl_wallet.create(contract_id, data, user_id, access_token)
        return self.write(ret)

    @myweb.authenticated
    @gen.coroutine
    def delete(self, contract_id, wallet_id):
        user_id = self.request.headers.get('x-user-id')
        access_token = self.request.headers.get('x-access-token')

        ret = yield ctrl_wallet.remove(contract_id, wallet_id, user_id, access_token)
        return self.write(ret)


class BalanceHandler(base_handler.BaseHandler):
    """
    钱包余额
    """

    @myweb.authenticated
    @gen.coroutine
    def get(self, contract_id, address):
        ret = yield ctrl_wallet.balance(contract_id, address)
        return self.write(ret)


class TransferHandler(base_handler.BaseHandler):
    """
    交易转账（含合约交易）
    """

    @myweb.authenticated
    @gen.coroutine
    def post(self, contract_id, from_address):
        user_id = self.request.headers.get('x-user-id')
        access_token = self.request.headers.get('x-access-token')

        data = json.loads(self.request.body.decode('utf-8'))
        ret = yield ctrl_wallet.transfer(contract_id, from_address, data, user_id, access_token)
        return self.write(ret)


class TxInfoHandler(base_handler.BaseHandler):
    """
    获取交易详情（含合约交易）
    """

    @myweb.authenticated
    @gen.coroutine
    def get(self, cid, tx_hash):
        user_id = self.request.headers.get('x-user-id')
        access_token = self.request.headers.get('x-access-token')

        ret = yield ctrl_wallet.get_tx_info(cid, tx_hash, user_id, access_token)
        return self.write(ret)



# coding=utf-8

import config
from qccwallet.http_handler import wallet_handler as w, contract_handler as c

base = '{}/{}/wallet'.format(config.VER, config.PLATFORM)
routes = [
    # 查询我的钱包账户列表
    (rf"/{base}/mine", w.MyWalletsHandler),

    # ******************************************************
    # 以下catalog/([a-f0-9]*)的含义是指在哪个区块链下，如eth，btc，eos等等，参数为注册时生成的类型ID
    # ******************************************************

    # *************************
    # 合约相关(注意：主币也是一种合约！)

    # 查询已注册合约列表
    (rf"/{base}/catalog/([a-f0-9]*)/contracts", c.QueryContractsHandler),

    # 向平台注册/注销合约（仅管理员可以使用）
    (rf"/{base}/catalog/([a-f0-9]*)/contract/([a-f0-9]*)", c.ContractHandler),

    # *************************
    # 基础账户/钱包相关

    # 创建/删除新钱包账户，含合约钱包，只是对应的contract_id不同
    (rf"/{base}/contract/([a-f0-9]*)/wallet/([a-f0-9]*)", w.WalletHandler),

    # 查询余额(根据钱包地址)
    (rf"/{base}/contract/([a-f0-9]*)/address/([a-fA-F0-9x]*)/balance", w.BalanceHandler),

    # 执行转账交易
    (rf"/{base}/contract/([a-f0-9]*)/address/([a-fA-F0-9x]*)/transfer", w.TransferHandler),

    # 查询交易详情
    (rf"/{base}/catalog/([a-f0-9]*)/txid/([a-fA-F0-9x]*)", w.TxInfoHandler),
]



# coding:utf-8

from qccwallet.service import ctrl_wallet
from asyncio import get_event_loop


async def query_mine(*auth_args):
    args = auth_args
    return await get_event_loop().run_in_executor(None, ctrl_wallet.query_mine, *args)


async def create(contract_id, data, *auth_args):
    args = (contract_id, data, *auth_args)
    return await get_event_loop().run_in_executor(None, ctrl_wallet.create, *args)


async def remove(contract_id, wallet_id, *auth_args):
    args = (contract_id, wallet_id, *auth_args)
    return await get_event_loop().run_in_executor(None, ctrl_wallet.remove, *args)


async def balance(contract_id, address):
    args = contract_id, address
    return await get_event_loop().run_in_executor(None, ctrl_wallet.balance, *args)


async def transfer(contract_id, from_address, data, *auth_args):
    args = contract_id, from_address, data, *auth_args
    return await get_event_loop().run_in_executor(None, ctrl_wallet.transfer, *args)


async def get_tx_info(cid, tx_hash, *auth_args):
    args = cid, tx_hash, *auth_args
    return await get_event_loop().run_in_executor(None, ctrl_wallet.get_tx_info, *args)
